/*		Ladereger fuer Bleiakkus

		Pin-Konfiguration
		PB0		TWI (I2C)-SDA
		PB1		Ausgangspin N-MOSFET zum Solar-Panel-Ueberbrueckung
		PB2		TWI (I2C)-SCL
		PB3		analoger Eingangspin Spannungsmessung
		PB4		Ausgangspin P-MOSFET zur Lastabschaltung

Die Abschaltung des Solarpanels erfolgt mit einem N-Kanal MOSFET durch kurzschliessen. Das
heisst bei High am Ausgang des AVR wird das Panel kurzgeschlossen und das Laden beendet.

Die Last Zu- und Abschaltung erfolgt mit einem P-Kanal MOSFET im Plus der
Versorgungsspannung. Das heisst, bei High am Ausgang des AVR wird die Last ab-
und bei low eingeschaltet.

Die Daten auf dem I2C-Bus werden wie folgt abgelegt:
										Adresse
Skalierter Istwert				LSB		0
Skalierter Istwert				MSB		1
Rohwert vom ADC					LSB		2
Rohwert vom ADC					MSB		3
Zaehler Ladeabschaltung 		LSB		4
Zaehler Ladeabschaltung 		MSB		5
Lastabschaltung (scaliert)		LSB		6
Lastabschaltung (scaliert)		MSB		7
Lasteinschaltung (scaliert)		LSB		8
Lasteinschaltung (scaliert)		MSB		9
Ladeabschaltung (scaliert)		LSB		10
Ladeabschaltung (scaliert)		MSB		11
Ladeeinschaltung (scaliert)		LSB		12
Ladeeinschaltung (scaliert)		MSB		13
*/
#include "config.h"
#include <stdlib.h>
#include <avr/io.h>
#include <inttypes.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <limits.h>
#include "adc.h"
#include "usi_twi_slave.h"
#include "watchdog.h"

#define TWI_ADDRESS		34				/**< I2C Adresse les Ladereglers					*/

#define ADC_LOAD_OFF_VOLT	1120L		/**< 11,20V Entladeschlusspannung Last abschalten	*/
#define ADC_LOAD_ON_VOLT	1165L		/**< 11.65V Last wieder einschalten					*/
#define ADC_CHARGE_OFF_VOLT	1380L		/**< 13.8V Ladeschlussspannung Solar kurzschliessen	*/
//#define ADC_CHARGE_ON_VOLT	1374L		/**< 13.74V Laden wieder einschalten				*/
#define ADC_CHARGE_ON_VOLT	1330L		/**< 13.30V Laden wieder einschalten				*/

#define ADC_MAX			16383			/**< maximaler Ausgabewert bei 5V Eingangsspannung	*/
#define ADC_INP_MAX		5				/**< max Spannung am analogen Eingang fuer ADC_MAX	*/
#define ADC_FACTOR		3				/**< Einstellung Spannungsteiler 15V Batterie ==> 5V*/
#define ADC_VALUE_MAX	(ADC_INP_MAX * ADC_FACTOR)	/**< realer Messwert der ADC_MAX entspricht	*/
#define ADC_COMMA		100				/**< Faktor zum Komma verschieben					*/

#define	ADC_DIV			(ADC_MAX + (ADC_COMMA / 2)) / ADC_COMMA
#define ADC_SCALE(value)	((((uint32_t)(ADC_VALUE_MAX) * (value) * 10) / (ADC_DIV) + 5) / 10)

#define INP_ADC_U			(1 << PB3)	/**< Eingangspin fuer Spannungsmessung				*/

#define OUTP_FET_CHARGE		(1 << PB1)	/**< Ausgangspin mit MOSFET zum Panel ueberbruecken	*/
#define OUTP_FET_LOAD		(1 << PB4)	/**< Ausgangspin mit MOSFET zur Lastabschaltung		*/

#define CHARGE_COUNTER_MAX	UINT16_MAX			/**< Maximalwert Ladeabschaltungs Zaehler	*/
#define CHARGE_COUNTER_INIT	0					/**< Initialisierungswert Ladeabschaltungs Zaehler	*/

#define	STATE_CHARGE		(1 << 0)	/**< Akku wird geladen								*/
#define	STATE_LOAD_OFF		(1 << 1)	/**< Last wurde abgeschaltet						*/


int main(void)
{
	uint8_t		state = 0;
	uint16_t	adc;		// ADC-Maximum generell:16384
	uint16_t	u_100;		//Spannung multipliziert mit 100
	uint16_t	charge_count = 0;		//Ladecounter

//Einganege festlegen (falls irgendwie nach Reset als Ausgang eingestellt)
	DDRB &= ~(INP_ADC_U);

//Ausgaenge festlegen
	DDRB |= OUTP_FET_CHARGE | OUTP_FET_LOAD;
//Ausgaenge auf 0 schreiben
	PORTB &= ~(OUTP_FET_CHARGE | OUTP_FET_LOAD);

	usi_twi_slave_init(TWI_ADDRESS);	// TWI slave init

# if 1
	WDT_Initialization_as_Wakeup_Source();
#endif
	sei();

/*	Schaltwerte abholbereit fuer den Master bereitlegen	*/
/*	Bereich fuer variable Werte initialisieren */
	twi_txbuffer[0]	= 0;
	twi_txbuffer[1]	= 0;
	twi_txbuffer[2]	= 0;
	twi_txbuffer[3]	= 0;
	twi_txbuffer[4]	= 0;
	twi_txbuffer[5]	= 0;
/*	Last aus	*/
	twi_txbuffer[6] = LOW_BYTE(ADC_LOAD_OFF_VOLT);
	twi_txbuffer[7] = HIGH_BYTE(ADC_LOAD_OFF_VOLT);
/*	Last ein	*/
	twi_txbuffer[8] = LOW_BYTE(ADC_LOAD_ON_VOLT);
	twi_txbuffer[9] = HIGH_BYTE(ADC_LOAD_ON_VOLT);
/*	Laden ein	*/
	twi_txbuffer[10] = LOW_BYTE(ADC_CHARGE_ON_VOLT);
	twi_txbuffer[11] = HIGH_BYTE(ADC_CHARGE_ON_VOLT);
/*	Laden aus	*/
	twi_txbuffer[12] = LOW_BYTE(ADC_CHARGE_OFF_VOLT);
	twi_txbuffer[13] = HIGH_BYTE(ADC_CHARGE_OFF_VOLT);

	while (1) {

		adc = adc_read(3);        // fuer PB3 14bit-Wert durch Oversampling
		u_100 = ADC_SCALE(adc);

		if (watchdog_flag() && charge_count < CHARGE_COUNTER_MAX)
			++charge_count;
// Last: wird per P-Kanal-MOSFET geschaltet, Pin TRUE ==> Last ausgeschaltet
// Bei Unterschreiten der Entladeschlusspannung Pin ein und damit FET2 aus
		if(u_100 < ADC_LOAD_OFF_VOLT) {
			PORTB |= OUTP_FET_LOAD;			//FET2 aus
			state |= STATE_LOAD_OFF;
		}
// Bei ueberschreiten der Entladeschlusspannung Pin aus und damit FET2 ein
		else if(u_100 > ADC_LOAD_ON_VOLT) {
			PORTB &= ~OUTP_FET_LOAD; 		//FET2 ein
			state &= ~STATE_LOAD_OFF;
		}

// Laden: FET schliesst mit TRUE Solarpanel kurz ==> kein Laden
// Bei unterschreiten der Einschaltschwelle FET ausschalten
		if(u_100 < ADC_CHARGE_ON_VOLT) {
			PORTB &= ~OUTP_FET_CHARGE;		//FET1 aus
			state |= STATE_CHARGE;
		}
// Bei ueberschreiten der Ladeschlussspannung: FET1 ein
		else if(u_100 > ADC_CHARGE_OFF_VOLT) {
			PORTB |= OUTP_FET_CHARGE;		//FET1 an
			state &= ~STATE_CHARGE;
		}

// Ist kein Laden aktiv Ladecounter auf Initialwert halten
		if (!(state & STATE_CHARGE))
			charge_count = CHARGE_COUNTER_INIT;

/*	skalierten Spannungswert in TWI-Puffer eintragen	*/
		twi_txbuffer[0]	= LOW_BYTE(u_100);
		twi_txbuffer[1]	= HIGH_BYTE(u_100);

/*	Rohwert des ADC in TWI-Puffer eintragen	*/
		twi_txbuffer[2] = LOW_BYTE(adc);
		twi_txbuffer[3] = HIGH_BYTE(adc);

/*	Ladeabschaltungszahler in TWI-Puffer eintragen	*/
		twi_txbuffer[4] = LOW_BYTE(charge_count);
		twi_txbuffer[5] = HIGH_BYTE(charge_count);

/* Eine Watchdog-Runde (8 sec) schlafen */
		wdt_reset();					// Watchdog zuruecksetzen
		WDTCR |= (1<<WDIE);				// Watchdog: Interrupt-Modus aktivieren
		MCUCR |= (1<<SM1) | (1<<SE);	// Power-Down aktivieren
		sleep_mode();

	}

return 0;
}
